@@ -5,213 +5,14 @@ import re
 import glob
 import importlib
 import inspect
-from functools import partial
 
 from PySide2 import QtCore, QtGui, QtWidgets
 from PySide2.QtUiTools import QUiLoader
 
-from pythontk import File, Str, Iter, hasAttribute, setAttributes
+from pythontk import File, Str, Iter, hasAttribute, setAttributes, formatReturn
 from uitk.styleSheet import StyleSheet
 
 
-class MainWindow(QtWidgets.QMainWindow):
-	def __init__(self, switchboard_instance, file):
-		'''Represents a main window in a GUI application.
-		Inherits from QtWidgets.QMainWindow class, providing additional functionality for 
-		managing user interface (UI) elements.
-
-		Parameters:
-			switchboard_instance (obj): An instance of the switchboard class
-			file_path (str): The full path to the UI file
-
-		Properties:
-			<UI>.name (str): The UI filename
-			<UI>.base (str): The base UI name
-			<UI>.path (str): The directory path containing the UI file
-			<UI>.tags (list): Any UI tags as a list
-			<UI>.level (int): The UI level
-			<UI>.isCurrentUi (bool): True if the UI is set as current
-			<UI>.isSubmenu (bool): True if the UI is a submenu
-			<UI>.isInitialized (bool): True after the UI is first shown
-			<UI>.connect (bool): Connect the UI's signals to it's slots.
-			<UI>.isConnected (bool): True if the UI is connected to its slots
-			<UI>.connectOnShow (bool): While True, the UI will be set as current on show
-			<UI>.preventHide (bool): While True, the hide method is disabled
-			<UI>.widgets (list): All the widgets of the UI
-			<UI>.slots (obj): The slots class instance
-		'''
-		super().__init__()
-
-		self.sb = switchboard_instance
-		self.name = File.formatPath(file, 'name')
-		setattr(self.sb, self.name, self)
-
-		self.path = File.formatPath(file, 'path')
-		self.level = self.sb._getUiLevelFromDir(file)
-		self.isSubmenu = self.level==2
-		self.isInitialized = False
-		self.isConnected = False
-		self.preventHide = False
-		self.connectOnShow = True
-		self.base = next(iter(self.name.split('_')))
-		self.tags = self.name.split('#')[1:]
-		self._widgets = set()
-		self._deferred = {}
-
-		ui = self.sb.load(file)
-		self.setWindowFlags(ui.windowFlags())
-		self.setCentralWidget(ui.centralWidget())
-
-		meta_obj = ui.metaObject() #transfer all property values from the dynamic ui.
-		for idx in range(meta_obj.propertyCount()):
-			prop = meta_obj.property(idx)
-			attr_name = prop.name()
-			value = ui.property(attr_name)
-			self.setProperty(attr_name, value)
-
-		if self.level>2:
-			self.sb.setStyle(self, style=self.sb.style)
-
-
-	def __getattr__(self, attr_name):
-		"""Looks for the widget in the parent class.
-		If found, the widget is initialized and returned, else an AttributeError is raised.
-
-		Parameters:
-			attr_name (str): the name of the attribute being accessed.
-
-		Return:
-			() The value of the widget attribute if it exists, or raises an AttributeError
-			if the attribute cannot be found.
-  
-		Raises:
-			AttributeError: if the attribute does not exist in the current instance
-			or the parent class.
-		"""
-		found_widget = self.sb._getWidgetFromUi(self, attr_name)
-		if found_widget:
-			self.sb.initWidgets(self, found_widget)
-			return found_widget
-
-		raise AttributeError(f'{self.__class__.__name__} has no attribute `{attr_name}`')
-
-
-	def event(self, event):
-		"""Handles events that are sent to the widget.
-
-		Parameters:
-			event (QtCore.QEvent): The event that was sent to the widget.
-
-		Return:
-			bool: True if the event was handled, otherwise False.
-
-		Notes:
-			This method is called automatically by Qt when an event is sent to the widget.
-			If the event is a `QEvent.ChildPolished` event, it calls the `on_child_polished`
-			method with the child widget as an argument. Otherwise, it calls the superclass
-			implementation of `event`.
-		"""
-		if event.type() == QtCore.QEvent.ChildPolished:
-			child = event.child()
-			self.on_child_polished(child)
-		return super().event(event)
-
-
-	def defer(self, func, *args, priority=0):
-		"""Defer execution of a function until later. The function is added to a dictionary of deferred 
-		methods, with a specified priority. Lower priority values will be executed before higher ones.
-		
-		Parameters:
-			func (function): The function to defer.
-			*args: Any arguments to be passed to the function.
-			priority (int, optional): The priority of the deferred method. Lower values will be executed 
-					first. Defaults to 0.
-		"""
-		method = partial(func, *args)
-		if priority in self._deferred:
-			self._deferred[priority] += (method,)
-		else:
-			self._deferred[priority] = (method,)
-
-
-	def trigger_deferred(self):
-		"""Executes all deferred methods, in priority order. Any arguments passed to the deferred functions
-		will be applied at this point. Once all deferred methods have executed, the dictionary is cleared.
-		"""
-		for priority in sorted(self._deferred):
-			for method in self._deferred[priority]:
-				method()
-		self._deferred.clear()
-
-
-	@property
-	def widgets(self):
-		"""Returns a list of the widgets in the widget's widget dictionary or initializes the widget dictionary and returns all the widgets found in the widget's children.
-
-		Return:
-			set: A set of the widgets in the widget's widget dictionary or all the widgets found in the widget's children.
-		"""
-		return self._widgets or self.sb.initWidgets(self, self.findChildren(QtWidgets.QWidget), returnAllWidgets=True)
-
-
-	@property
-	def slots(self):
-		"""Returns a list of the slots connected to the widget's signals.
-
-		Return:
-			list: A list of the slots connected to the widget's signals.
-		"""
-		return self.sb.getSlots(self)
-
-
-	@property
-	def isCurrentUi(self):
-		"""Returns True if the widget is the currently active UI, False otherwise."""
-		return self==self.sb.getCurrentUi()
-
-
-	def setAsCurrent(self):
-		"""Sets the widget as the currently active UI."""
-		self.sb.setCurrentUi(self)
-
-
-	def connect(self):
-		"""Connects the widget's signals to their respective slots."""
-		self.sb.setConnections(self)
-
-
-	def on_child_polished(self, w):
-		"""Called after a child widget is polished. Initializes the widget dictionary with the child widget and connects its signals to their respective slots if the widget is connected.
-
-		Parameters:
-			w (QWidget): The polished child widget.
-		"""
-		if w not in self._widgets:
-			self.sb.initWidgets(self, w)
-			self.trigger_deferred()
-			if self.isConnected:
-				self.sb.connectSlots(self, w)
-
-
-	def setVisible(self, state):
-		"""Called every time the widget is shown or hidden on screen. If the widget is set to be prevented from hiding, it will not be hidden when state is False.
-
-		Parameters:
-			state (bool): Whether the widget is being shown or hidden.
-		"""
-		if state: #visible
-			if self.connectOnShow and not self.isConnected:
-				self.sb.setConnections(self)
-			self.activateWindow()
-			# self.raise_()
-			self.setWindowFlags(self.windowFlags()|QtCore.Qt.WindowStaysOnTopHint)
-			self.isInitialized = True
-		elif self.preventHide: #invisible
-			return
-		super().setVisible(state)
-
-
-
 class Switchboard(QUiLoader, StyleSheet):
 	"""Load dynamic UI, assign convenience properties, and handle slot connections.
 
@@ -220,6 +21,7 @@ class Switchboard(QUiLoader, StyleSheet):
 		sb.ui: Returns the current UI.
 		sb.<uiFileName>: Accesses the UI loaded from uiFileName.
 		sb.<customWidgetClassName>: Accesses the custom widget with the specified class name.
+		sb.<slotsClassName>: Accesses the slots class of the specified class name.
 
 	Parameters:
 		parent (obj): A QtObject derived class.
@@ -715,7 +517,10 @@ class Switchboard(QUiLoader, StyleSheet):
 
 
 	def setSlots(self, ui, clss):
-		'''Explicitly set the slot class for a given ui.
+		'''This function sets the slot class for a loaded dynamic UI object. It takes a UI object and 
+		a class or class instance and sets the class as the slots location for the given UI. It then 
+		initializes the widgets and returns the slot class instance. The slots class is given an 
+		attribute `get_switchboard_instance` which can be called to retrieve it's switchboard instance.
 
 		Parameters:
 			ui (obj): A previously loaded dynamic ui object.
@@ -724,20 +529,26 @@ class Switchboard(QUiLoader, StyleSheet):
 		Return:
 			(obj) class instance.
 		'''
-		if inspect.isclass(clss):
-			setattr(clss, 'get_switchboard_instance', lambda slots_inst: self)
-			try:
-				ui._slots = clss() if callable(clss) else clss
-			except AttributeError as error:
-				print (f'# Error: {__file__} in setSlots\n#\t{ui.name} {clss} {error}')
-				ui._slots = None
-			self.initWidgets(ui, ui.findChildren(QtWidgets.QWidget))
-			return ui._slots
-		return None
+		if not inspect.isclass(clss):
+			return None
+
+		setattr(clss, 'get_switchboard_instance', lambda slots_inst: self) #set an attribute for the slots class that returns it's switchboard instance. ie. <Your_slots_class>.get_switchboard_instance()
+
+		try:
+			ui._slots = clss() if callable(clss) else clss
+			setattr(self, ui._slots.__class__.__name__, ui._slots) #set the slots class name as a switchboard attribute. ie. <switchboard>.Your_slots_class_name
+		except Exception as error:
+			print (f'# Error: {__file__} in setSlots\n#\t{error} in {clss}')
+			ui._slots = None
+
+		self.initWidgets(ui, ui.findChildren(QtWidgets.QWidget))
+		return ui._slots
 
 
 	def getSlots(self, ui, persist=False):
-		'''Get the class instance of the ui's slots module if it exists.
+		'''This function tries to get a class instance of the slots module from a dynamic ui object. 
+		If it doesn't exist, it tries to import it from a specified location or from the parent menu. 
+		If it's found, it's set and returned, otherwise None is returned.
 
 		Parameters:
 			ui (obj): A previously loaded dynamic ui object.
@@ -856,7 +667,7 @@ class Switchboard(QUiLoader, StyleSheet):
 			except Exception as error:
 				print (f'# Error: {__file__} in registerWidgets\n#\t{error}.')
 
-		return Iter.formatReturn(result, widgets) #if 'widgets' is given as a list; return a list.
+		return formatReturn(result, widgets) #if 'widgets' is given as a list; return a list.
 
 
 	def loadAllUi(self, path=None, widgets=None):
@@ -907,7 +718,7 @@ class Switchboard(QUiLoader, StyleSheet):
 				fullpath = os.path.join(self.widgets_location, mod_name+'.py')
 				self.registerWidgets(fullpath)
 
-		ui = MainWindow(self, file)
+		ui = self.MainWindow(self, file)
 		self._loadedUi.add(ui)
 		return ui
 
@@ -922,40 +733,11 @@ class Switchboard(QUiLoader, StyleSheet):
 		Return:
 			(str)(list) list if 'level' given as a list.
 		'''
-		error_msg = lambda ui: f'# Error: {__file__} in getUi\n#\tUI not found: {ui}({type(ui).__name__})\n#\tConfirm the following UI path is correct: {self.ui_location}'
-
 		if not ui:
 			ui = self.ui
 
 		if level:
-			if isinstance(ui, str):
-				ui, *tags = ui.split('#')
-
-				ui1 = self.getUi(ui)
-				if not ui1 or level==2:
-					submenu_name = '{}_{}#{}'.format(ui, 'submenu', '#'.join(tags)).rstrip('#') #reformat as submenu w/tags. ie. 'polygons_submenu#edge' from 'polygons'
-					ui1 = self.getUi(submenu_name) #in the case where a submenu exist without a parent menu.
-					if not ui1:
-						print (error_msg(ui))
-						return None
-
-				ui = [u for u in self._loadedUi
-						if all([
-							u.base==ui1.base, 
-							u.tags==Iter.makeList(tags), 
-							u.level in Iter.makeList(level),
-						])
-				]
-
-			else:
-				ui = [u for u in self._loadedUi 
-						if all([
-							u.base==ui.base, 
-							u.level in Iter.makeList(level),
-						])
-				]
-
-			return Iter.formatReturn(ui, level) #if 'level' is given as a list; return a list.
+			return self._getUiAtLevel(ui, level)
 
 		elif isinstance(ui, str):
 			try:
@@ -965,8 +747,7 @@ class Switchboard(QUiLoader, StyleSheet):
 				return getattr(self, ui)
 
 			except AttributeError as error:
-				print (error_msg(ui))
-				return None
+				return print (f'# Error: {__file__} in getUi\n#\tUI not found: {ui}({type(ui).__name__})\n#\tConfirm the following UI path is correct: {self.ui_location}')
 
 		elif isinstance(ui, (list, set, tuple)):
 			return [self.getUi(u) for u in ui]
@@ -975,6 +756,45 @@ class Switchboard(QUiLoader, StyleSheet):
 			return ui
 
 
+	def _getUiAtLevel(self, ui, level):
+		'''Extends `getUi` to get the ui(s) at the given level(s).
+
+		Parameters:
+			ui (str)(obj)(list): The ui or name(s) of the ui.
+			level (int)(list): Integer(s) representing the level to include.
+						ex. 2 for submenu, 3 for main menu, or [2, 3] for both.
+		Return:
+			(str)(list) list if 'level' given as a list.
+		'''
+		if isinstance(ui, str):
+			ui, *tags = ui.split('#')
+
+			ui1 = self.getUi(ui)
+			if not ui1 or level==2:
+				submenu_name = '{}_{}#{}'.format(ui, 'submenu', '#'.join(tags)).rstrip('#') #reformat as submenu w/tags. ie. 'polygons_submenu#edge' from 'polygons'
+				ui1 = self.getUi(submenu_name) #in the case where a submenu exist without a parent menu.
+				if not ui1:
+					return print (f'# Error: {__file__} in getUi\n#\tUI not found: {ui}({type(ui).__name__})\n#\tConfirm the following UI path is correct: {self.ui_location}')
+
+			ui = [u for u in self._loadedUi
+					if all([
+						u.base==ui1.base, 
+						u.tags==Iter.makeList(tags), 
+						u.level in Iter.makeList(level),
+					])
+			]
+
+		else:
+			ui = [u for u in self._loadedUi 
+					if all([
+						u.base==ui.base, 
+						u.level in Iter.makeList(level),
+					])
+			]
+
+		return formatReturn(ui, level) #if 'level' is given as a list; return a list.
+
+
 	def showAndConnect(self, ui, connectOnShow=True):
 		'''Register the uiName in history as current, 
 		set slot connections, and show the given ui.
@@ -1001,6 +821,7 @@ class Switchboard(QUiLoader, StyleSheet):
 		except AttributeError as error:
 			if len(self._loadedUi)==1: #if only one ui is loaded set that ui as current.
 				ui = self._loadedUi.pop()
+				self._loadedUi.add(ui)
 				self.setCurrentUi(ui)
 				return ui
 			elif self.ui_location.endswith('.ui'): #if the ui location is set to a single ui, then load and set that ui as current.
@@ -1896,18 +1717,18 @@ if __name__=='__main__':
 
 	app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv) #return the existing QApplication object, or create a new one if none exists.
 
-	from uitk.slots.polygons import Polygons
-	sb = Switchboard(slots_location='slots/maya') #set relative paths, and explicity set the slots class instead of providing a path like: slots='slots/maya', which in this case would produce the same result with just a little more overhead.
-	ui = sb.polygons #get the ui by it's name.
+	from uitk.slots.example import Example
+	sb = Switchboard(slots_location=Example) #set relative paths, and explicity set the slots class instead of providing a path like: slots='slots/maya', which in this case would produce the same result with just a little more overhead.
+	ui = sb.example #get the ui by it's name.
 	ui.show()
 
-	print ('current ui:', sb.ui)
-	print ('current ui name:', sb.ui.name)
-	print ('connected state:', ui.isConnected)
-	print ('slots:', ui.slots)
-	print ('method tb000:', ui.tb000.getSlot())
-	print ('widget from method tb000:', sb.getWidgetFromMethod(ui.tb000.getSlot()))
-	# print ('widgets:', ui.widgets)
+	print ('ui:'.ljust(25), sb.ui)
+	print ('ui name:'.ljust(25), sb.ui.name)
+	print ('connected state:'.ljust(25), ui.isConnected)
+	print ('slots:'.ljust(25), ui.slots)
+	print ('method tb000:'.ljust(25), ui.tb000.getSlot())
+	print ('widget from method tb000:'.ljust(25), sb.getWidgetFromMethod(ui.tb000.getSlot()))
+	# for w in ui.widgets: print ('child widget:'.ljust(25), (w.name or type(w).__name__).ljust(25), w.prefix.ljust(25), w.type.ljust(15), w.derivedType.ljust(15), id(w))
 
 	exit_code = app.exec_()
 	if exit_code != -1:
